__all__ = ()

CO_OPTIMIZED = 1
CO_NEWLOCALS = 2
CO_VARARGS = 4
CO_VARKEYWORDS = 8
CO_NESTED = 16
CO_GENERATOR = 32
CO_NOFREE = 64
    
CO_COROUTINE = 128
CO_ITERABLE_COROUTINE = 256
CO_ASYNC_GENERATOR = 512
# matches `async def` functions and `@coroutine` functions.
CO_COROUTINE_ALL = CO_COROUTINE | CO_ITERABLE_COROUTINE
